#include "WavFile.h"
#include "Error.h"

WavFile::WavFile()
{
    mode = UNDEFINED_MODE;
    unsigned char tmpWavHeader[44] = {'R','I','F','F',0x80,0x80,0x80,0x80,'W','A','V','E','f','m','t',' ',0x10,0,0,0,1,0,1,0,0x22,0x56,0,0,0x22,0x56,0,0,1,0,8,0,'d','a','t','a',0,0,0,0 };
    memcpy(wavHeader, tmpWavHeader, 44);
    volume = 100;
}

WavFile::~WavFile()
{
}

int WavFile::openInput(char *filename)
{
    if (mode == UNDEFINED_MODE)
        mode = INPUT_MODE;
    else
    {
        cerr << "Wav file already open";
        exit (1);
    }
	ifsWav.open(filename,ios::in|ios::binary|ios::ate);//|ios::nocreate);
	if (ifsWav == 0) {
		cerr << "Unable to open " << filename << "\n";
		exit(1);
	}
	WavFileName=filename;
	fileSize=ifsWav.tellg();
	filePosition = 0;
	blockNumber = 0;
	bufferPosition = 0;
    try
    {
	    getBlock(blockNumber);

	    format = wordVal(WAV_HDR_FORMAT);
	    channels = wordVal(WAV_HDR_CHANNELS);
	    sampleRate = longWordVal(WAV_HDR_SAMPLE_RATE);
	    bytesPerSecond = longWordVal(WAV_HDR_BYTES_PER_SEC);
        bytesPerSample = wordVal(WAV_HDR_BYTES_PER_SAMPLE);
        bitsPerSample = wordVal(WAV_HDR_BITS_PER_SAMPLE);
	    length = longWordVal(WAV_HDR_LENGTH);
	    time = length/bytesPerSecond;
	    ifsWav.seekg(0);

        if (verbosity > 1)
            describeWav();
    }

    catch (Error::endOfFile)
    {
        cerr << "Unexpected end of .wav file, unable to get .wav file information!\n";
        exit(1);
    }

    return 0;
}

int WavFile::openOutput(char *pFileName, long pSampleRate, int pBitsPerSample, int pChannels)
{
    if (mode == UNDEFINED_MODE)
        mode = OUTPUT_MODE;
    else
    {
        cerr << "Wav file already open for output\n";
        exit(1);
    }

    ofsWav.open(pFileName,ios::out|ios::binary|ios::ate|ios::trunc);

    if (ofsWav == 0) {
		cerr << "Unable to open " << pFileName << "\n";
        return 0;
		exit(1);
	}

    if (pBitsPerSample != 8 && pBitsPerSample != 16)
    {
        cerr << "Only 8 and 16 bits per sample supported\n";
        exit(1);
    }

	if (pSampleRate != 22050 && pSampleRate != 44100)
	{
		cerr << "Only 22050 and 44100 sample rates supported\n";
		exit(1);
	}

	if (pChannels != 1 && pChannels != 2)
	{
		cerr << "Only 1 and 2 channel .wav files supported\n";
		exit(1);
	}

    format = 0x01;
    channels = pChannels;
    sampleRate = pSampleRate;
    bitsPerSample = pBitsPerSample;
    bytesPerSample = bitsPerSample / 8 * channels;
    bytesPerSecond = sampleRate * bytesPerSample;
    volume = 95;
    length = 0;
    time = 0;

    WordVal::putWord(wavHeader, WAV_HDR_FORMAT, format);
    WordVal::putWord(wavHeader, WAV_HDR_CHANNELS, channels);
    WordVal::putWord(wavHeader, WAV_HDR_SAMPLE_RATE, sampleRate);
    WordVal::putLongWord(wavHeader, WAV_HDR_BYTES_PER_SEC, bytesPerSecond);
    WordVal::putWord(wavHeader, WAV_HDR_BYTES_PER_SAMPLE, bytesPerSample);
    WordVal::putWord(wavHeader, WAV_HDR_BITS_PER_SAMPLE, bitsPerSample);
    WordVal::putWord(wavHeader, WAV_HDR_LENGTH, length);

    ofsWav.write((char *)wavHeader,44);

	return 0;
}

void WavFile::closeOutput()
{
    if (!ofsWav.is_open())
        return;
    if (mode != OUTPUT_MODE)
    {
        cerr << "Can not close output wav file\n";
        return;
        exit(1);
    }
	long flength, wavFlength;
	unsigned char lengthLword[4];
	int i;

	wavFlength = (double)ofsWav.tellp() - (double)8;
	flength = wavFlength - 36;
	ofsWav.seekp(0x28,ios::beg);
    WordVal::decToLongWord(flength, lengthLword);

    for(i=0;i<4;i++)
		ofsWav.put((unsigned char)lengthLword[i]);

    /* The file length for the RIFF chunk of the .wav file is in wavFlength */
    /* Write the file length in lo-byte/hi-byte, lo-word/hi-word format to the .wav file */

	ofsWav.seekp(0x04,ios::beg);
    WordVal::decToLongWord(wavFlength, lengthLword);

    for(i=0;i<4;i++)
		ofsWav.put((unsigned char)lengthLword[i]);

    ofsWav.close();
    mode = UNDEFINED_MODE;
}

double WavFile::wordVal(double WordFilePosition)
{
    getWord(WordFilePosition, word);
    return WordVal::wordVal((unsigned char*)word);
}

double WavFile::signedWordVal(double WordFilePosition)
{
    getWord(WordFilePosition, word);
    return WordVal::signedWordVal((unsigned char*)word);
}

double WavFile::longWordVal(double longWordFilePosition)
{
    getWord(longWordFilePosition, longWord);
    getWord(longWordFilePosition+2, longWord+2);
    return WordVal::longWordVal((unsigned char*)longWord);
}

size_t WavFile::getBlock(long block_number)
{
    if (mode != INPUT_MODE)
    {
        cerr << "Can not read from unopened input file\n";
        exit(1);
    }
	ifsWav.seekg(block_number * BLOCKSIZE);
	bufferPosition = 0;
	blockNumber = block_number;
	ifsWav.read((char *)buffer,BLOCKSIZE);
	if (ifsWav.gcount() < 1)
    {
		endOfFile();
	}
	return 0;
}

void WavFile::endOfFile()
{
    ifsWav.close();
	throw Error::endOfFile();
}

unsigned int WavFile::getByte(double byteFilePosition)
{
	long newBlockNumber;

    if (ifsWav.is_open())
    {
        if (mode != INPUT_MODE)
        {
            cerr << "Can not read from unopened input file\n";
            exit(1);
        }
        if (byteFilePosition >= fileSize)
        {
			endOfFile();
	    }
	    else
        {
		    newBlockNumber = (long)floor(byteFilePosition/BLOCKSIZE);
		    if (newBlockNumber != blockNumber)
            {
			    getBlock(newBlockNumber);
		    }
		    bufferPosition =
			    (long)(byteFilePosition-(newBlockNumber*BLOCKSIZE));
		    byte = buffer[bufferPosition];
	    }
    }
    else
		endOfFile();
    return byte;
}

void WavFile::putSample(int pAmplitude)
{
    if (mode != OUTPUT_MODE)
    {
        cerr << "Can not write to .wav file\n";
        exit(1);
    }
    if (pAmplitude < -100)
        pAmplitude = -100;
    if (pAmplitude > 100)
        pAmplitude = 100;
    if (bitsPerSample == 8)
    {
        pAmplitude = ((pAmplitude + 100) * 0xff / 200);
        pAmplitude = (double)pAmplitude * (double)((double)volume / (double)100);
        ofsWav.put((unsigned char)pAmplitude);
    } else
    {
        unsigned char signedWord[2];
        double lAmplitude = ((pAmplitude + 100) / 200) * 0xffff;
        lAmplitude = ((double)lAmplitude * ((double)volume / (double)100)) - (double)0x8000;
        WordVal::decToSignedWord(lAmplitude, signedWord);
        ofsWav.put(signedWord[0]);
        ofsWav.put(signedWord[1]);
    }
}

int WavFile::getSample(void)
{
    int amplitude;
    if (mode != INPUT_MODE)
    {
        cerr << "Can not read from .wav file\n";
        exit(1);
    }
    if (bitsPerSample == 8)
    {
        unsigned int lAmplitude = getByte(filePosition++);
        if ((double)lAmplitude > (double)0x80)
            amplitude = (((double)lAmplitude - (double)0x80)/ (double)0x80) * 100;
        else
            amplitude = (((double)0x80 - (double)lAmplitude) / -(double)0x80) * 100;
        amplitude = (double)amplitude * (double)volume / double(100);
    } else
    {
        unsigned char lAmplitude[2];

        lAmplitude[0] = getByte(filePosition++);
        lAmplitude[1] = getByte(filePosition++);

        double rawAmplitude = WordVal::signedWordVal(lAmplitude);
        if (rawAmplitude >= 0)
            amplitude = int((rawAmplitude / (double)0x7fff) * 100);
        else
            amplitude = int((rawAmplitude / (double)0x8000) * 100);
    }
    if (amplitude < -100)
        amplitude = -100;
    if (amplitude > 100)
        amplitude = 100;
    return amplitude;
}

void WavFile::getWord(double byteFilePosition, unsigned char *word)
{
    word[0] = getByte(byteFilePosition);
    word[1] = getByte(byteFilePosition + 1);
}

bool WavFile::eof(void)
{
    validateMode();
    if (mode == INPUT_MODE)
    {
		if (filePosition >= fileSize)
            return true;
    }
    else if (mode == OUTPUT_MODE)
    {
        if (ofsWav.eof())
            return true;
    }
    return false;
}

void WavFile::describeWav(void)
{
    validateMode();
	cout << "format: " << format << '\n';
	cout << "channels: " << channels << '\n';
	cout << "Sample Rate: " << sampleRate << '\n';
	cout << "Bytes per Second: " << bytesPerSecond << '\n';
	cout << "Bytes per Sample: " << bytesPerSample << '\n';
	cout << "Bits per Sample: " << bitsPerSample << '\n';
	cout.setf(ios::fixed,ios::floatfield);
	cout.precision(0);
	cout << "length: " << length << '\n';
	cout.precision(2);
	cout << "time: "  << time << '\n';
}

double WavFile::tellp(void)
{
    validateMode();
    if (mode == OUTPUT_MODE)
        return ofsWav.tellp();
    return ifsWav.tellg();
}

void WavFile::validateMode(void)
{
    if (mode == UNDEFINED_MODE)
    {
        cerr << "Can not read from unopened input file\n";
        exit(1);
    }
}

void WavFile::putSilence(double silenceLength)
{
    for (int i=0; i<(int)(silenceLength * bytesPerSecond); i++)
        putSample(0);
}

WavTimer::WavTimer(WavFile *pWavFile)
{
    wavFile = pWavFile;
    reset();
}

void WavTimer::reset()
{
    if (wavFile->mode == INPUT_MODE)
        startByte = (streampos)wavFile->filePosition;
    else
        startByte = (streampos)wavFile->ofsWav.tellp();
}

void WavTimer::set(streampos pFileByte)
{
    startByte = pFileByte;
}

streampos WavTimer::getStartByte()
{
	return startByte;
}

double WavTimer::currentTime()
{
    if (wavFile->mode == INPUT_MODE)
        return ((wavFile->filePosition - startByte)+1)/wavFile->bytesPerSecond;
    else if (wavFile->mode == OUTPUT_MODE)
    {
        double j = (((double)wavFile->ofsWav.tellp() - (double)startByte)+(double)1)/(double)wavFile->bytesPerSecond;
        return j;
    }
    else
    {
        cerr << "Cannot show current time for Wav file before it is opened\n";
        return 0;
    }
}